package pwhash

import (
	"crypto/rand"
	"crypto/subtle"
	"encoding/hex"
	"errors"
	"fmt"
	"io"
	"strconv"
	"strings"

	"github.com/amoghe/go-crypt"
	"golang.org/x/crypto/argon2"
	"golang.org/x/crypto/scrypt"
)

// PasswordHash is the interface for a password hashing algorithm
// implementation.
type PasswordHash interface {
	// ComparePassword returns true if the given password matches
	// the encrypted one.
	ComparePassword(string, string) bool

	// Encrypt the given password.
	Encrypt(string) string
}

// systemCryptPasswordHash uses the glibc crypt function.
type systemCryptPasswordHash struct {
	hashStr string
}

// NewSystemCrypt returns a PasswordHash that uses the system crypt(3)
// function, specifically glibc with its SHA512 algorithm.
func NewSystemCrypt() PasswordHash {
	return &systemCryptPasswordHash{"$6$"}
}

// ComparePassword returns true if the given password matches the
// encrypted one.
func (s *systemCryptPasswordHash) ComparePassword(encrypted, password string) bool {
	enc2, err := crypt.Crypt(password, encrypted)
	if err != nil {
		return false
	}
	return subtle.ConstantTimeCompare([]byte(encrypted), []byte(enc2)) == 1
}

// Encrypt the given password using glibc crypt.
func (s *systemCryptPasswordHash) Encrypt(password string) string {
	salt := fmt.Sprintf("%s%x$", s.hashStr, getRandomBytes(16))
	enc, err := crypt.Crypt(password, salt)
	if err != nil {
		panic(err)
	}
	return enc
}

var (
	argonKeyLen  uint32 = 32
	argonSaltLen        = 16
)

// Argon2PasswordHash uses the Argon2 hashing algorithm.
type argon2PasswordHash struct {
	// Encoding parameters.
	params argon2Params
}

// NewArgon2 returns an Argon2i-based PasswordHash using the default parameters.
func NewArgon2() PasswordHash {
	return NewArgon2WithParams(
		defaultArgon2Params.Time,
		defaultArgon2Params.Memory,
		defaultArgon2Params.Threads,
	)
}

// NewArgon2WithParams returns an Argon2i-based PasswordHash using the
// specified parameters for time, memory, and number of threads.
func NewArgon2WithParams(time, mem uint32, threads uint8) PasswordHash {
	return &argon2PasswordHash{
		params: argon2Params{
			Time:    time,
			Memory:  mem,
			Threads: threads,
		},
	}
}

// ComparePassword returns true if the given password matches the
// encrypted one.
func (s *argon2PasswordHash) ComparePassword(encrypted, password string) bool {
	params, salt, dk, err := decodeArgon2Hash(encrypted)
	if err != nil {
		return false
	}
	dk2 := argon2.Key([]byte(password), salt, params.Time, params.Memory, params.Threads, argonKeyLen)
	return subtle.ConstantTimeCompare(dk, dk2) == 1
}

// Encrypt the given password with the Argon2 algorithm.
func (s *argon2PasswordHash) Encrypt(password string) string {
	salt := getRandomBytes(argonSaltLen)
	dk := argon2.Key([]byte(password), salt, s.params.Time, s.params.Memory, s.params.Threads, argonKeyLen)
	return encodeArgon2Hash(s.params, salt, dk)
}

type argon2Params struct {
	Time    uint32
	Memory  uint32
	Threads uint8
}

// Default Argon2 parameters are tuned for a high-traffic
// authentication service (<1ms per operation).
var defaultArgon2Params = argon2Params{
	Time:    1,
	Memory:  4 * 1024,
	Threads: 4,
}

func encodeArgon2Hash(params argon2Params, salt, dk []byte) string {
	return fmt.Sprintf("$a2$%d$%d$%d$%x$%x", params.Time, params.Memory, params.Threads, salt, dk)
}

func decodeArgon2Hash(s string) (params argon2Params, salt []byte, dk []byte, err error) {
	if !strings.HasPrefix(s, "$a2$") {
		err = errors.New("not an Argon2 password hash")
		return
	}

	parts := strings.SplitN(s[4:], "$", 5)
	if len(parts) != 5 {
		err = errors.New("bad encoding")
		return
	}

	var i uint64

	if i, err = strconv.ParseUint(parts[0], 10, 32); err != nil {
		return
	}
	params.Time = uint32(i)

	if i, err = strconv.ParseUint(parts[1], 10, 32); err != nil {
		return
	}
	params.Memory = uint32(i)

	if i, err = strconv.ParseUint(parts[2], 10, 8); err != nil {
		return
	}
	params.Threads = uint8(i)

	if salt, err = hex.DecodeString(parts[3]); err != nil {
		return
	}
	dk, err = hex.DecodeString(parts[4])
	return
}

var (
	scryptKeyLen  = 32
	scryptSaltLen = 16
)

// ScryptPasswordHash uses the scrypt hashing algorithm.
type scryptPasswordHash struct {
	params scryptParams
}

// NewScrypt returns a PasswordHash that uses the scrypt algorithm
// with the default parameters.
func NewScrypt() PasswordHash {
	return NewScryptWithParams(
		defaultScryptParams.N,
		defaultScryptParams.R,
		defaultScryptParams.P,
	)
}

// NewScryptWithParams returns a PasswordHash that uses the scrypt
// algorithm with the specified parameters.
func NewScryptWithParams(n, r, p int) PasswordHash {
	return &scryptPasswordHash{
		params: scryptParams{
			N: n,
			R: r,
			P: p,
		},
	}
}

// ComparePassword returns true if the given password matches
// the encrypted one.
func (s *scryptPasswordHash) ComparePassword(encrypted, password string) bool {
	params, salt, dk, err := decodeScryptHash(encrypted)
	if err != nil {
		return false
	}
	dk2, err := scrypt.Key([]byte(password), salt, params.N, params.R, params.P, scryptKeyLen)
	if err != nil {
		return false
	}
	return subtle.ConstantTimeCompare(dk, dk2) == 1
}

// Encrypt the given password with the scrypt algorithm.
func (s *scryptPasswordHash) Encrypt(password string) string {
	salt := getRandomBytes(scryptSaltLen)

	dk, err := scrypt.Key([]byte(password), salt, s.params.N, s.params.R, s.params.P, scryptKeyLen)
	if err != nil {
		panic(err)
	}

	return encodeScryptHash(s.params, salt, dk)
}

type scryptParams struct {
	N int
	R int
	P int
}

var defaultScryptParams = scryptParams{
	N: 16384,
	R: 8,
	P: 1,
}

func encodeScryptHash(params scryptParams, salt, dk []byte) string {
	return fmt.Sprintf("$s$%d$%d$%d$%x$%x", params.N, params.R, params.P, salt, dk)
}

func decodeScryptHash(s string) (params scryptParams, salt []byte, dk []byte, err error) {
	if !strings.HasPrefix(s, "$s$") {
		err = errors.New("not a scrypt password hash")
		return
	}

	parts := strings.SplitN(s[3:], "$", 5)
	if len(parts) != 5 {
		err = errors.New("bad encoding")
		return
	}

	if params.N, err = strconv.Atoi(parts[0]); err != nil {
		return
	}

	if params.R, err = strconv.Atoi(parts[1]); err != nil {
		return
	}
	if params.P, err = strconv.Atoi(parts[2]); err != nil {
		return
	}

	if salt, err = hex.DecodeString(parts[3]); err != nil {
		return
	}
	dk, err = hex.DecodeString(parts[4])
	return
}

func getRandomBytes(n int) []byte {
	b := make([]byte, n)
	_, err := io.ReadFull(rand.Reader, b[:])
	if err != nil {
		panic(err)
	}
	return b
}

// A registry of default handlers for decoding passwords.
var prefixRegistry = map[string]PasswordHash{
	"$1$":  NewSystemCrypt(),
	"$5$":  NewSystemCrypt(),
	"$6$":  NewSystemCrypt(),
	"$s$":  NewScrypt(),
	"$a2$": NewArgon2(),
}

// ComparePassword returns true if the given password matches the
// encrypted one.
func ComparePassword(encrypted, password string) bool {
	for pfx, h := range prefixRegistry {
		if strings.HasPrefix(encrypted, pfx) {
			return h.ComparePassword(encrypted, password)
		}
	}
	return false
}

// DefaultEncryptAlgorithm is used by the Encrypt function to encrypt
// passwords.
var DefaultEncryptAlgorithm PasswordHash

func init() {
	DefaultEncryptAlgorithm = NewArgon2()
}

// Encrypt will encrypt a password with the default algorithm.
func Encrypt(password string) string {
	return DefaultEncryptAlgorithm.Encrypt(password)
}
